from IPython.core import magic_arguments
from IPython.core.magic import Magics, cell_magic, magics_class, needs_local_scope

from ._connection import KamuConnection


@magics_class
class KamuMagics(Magics):
    """
    Kamu magics for Jupyter / IPython.
    """

    @cell_magic
    @needs_local_scope
    @magic_arguments.magic_arguments()
    @magic_arguments.argument(
        "-c",
        "--connection",
        type=str,
        default=None,
        help="Variable name that holds KamuConnection",
    )
    @magic_arguments.argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="If present, query will be stored in variable of this name.",
    )
    @magic_arguments.argument(
        "-q",
        "--quiet",
        type=bool,
        default=False,
        const=True,
        nargs="?",
        help="Execute query without returning the result dataframe.",
    )
    @magic_arguments.argument(
        "-n",
        "--max-rows",
        type=int,
        default=None,
        help="Maximum number of rows that will be pulled back from the server for SQL queries",
    )
    def sql(self, line, cell, local_ns=None):
        args = magic_arguments.parse_argstring(self.sql, line)

        # Get connection object
        if args.connection:
            connection = local_ns.get(args.connection)
        else:
            for var_name in ["con", "conn", "connection"]:
                connection = local_ns.get(var_name)
                if isinstance(connection, KamuConnection):
                    break

        if not connection:
            raise ValueError(
                "Active KamuConnection not found, please provide -c <connection> argument"
            )
        if not isinstance(connection, KamuConnection):
            raise ValueError(
                f"Expected instance of KamuConnection, but found: {repr(connection)}"
            )

        sql = cell.strip()
        df = connection.query(sql)

        if args.output:
            local_ns[args.output] = df

        if not args.quiet:
            return df



import adbc_driver_flightsql.dbapi
import adbc_driver_manager
import pandas

from ._connection import KamuConnection


class KamuConnectionFlightSql(KamuConnection):
    """
    `KamuConnection` implementation using ADBC client protocol
    """

    def __init__(self, url, token=None, **connection_params):
        super().__init__()

        self._url = url

        if token:
            connection_params[
                adbc_driver_flightsql.DatabaseOptions.AUTHORIZATION_HEADER.value
            ] = f"Bearer {token}"
        else:
            # We don't allow fully non-authenticated access.
            # Anonymous users still have to go through basic auth procedure
            # so that server can reject new client, apply backpressure, or assign
            # a temporary session token that can be used to identify the client and
            # apply individual rate limiting strategies.
            connection_params[adbc_driver_manager.DatabaseOptions.USERNAME.value] = (
                "anonymous"
            )
            connection_params[adbc_driver_manager.DatabaseOptions.PASSWORD.value] = ""

        self._adbc = adbc_driver_flightsql.dbapi.connect(
            self._url,
            db_kwargs=connection_params,
            autocommit=True,
        )

    def __repr__(self):
        return f"KamuConnectionFlightSql(url='{self._url}')"

    def url(self):
        return self._url

    def query(self, sql):
        return pandas.read_sql(sql, self.as_adbc())

    def as_adbc(self):
        return self._adbc

    def __enter__(self):
        self._adbc.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._adbc.__exit__(exc_type, exc_value, traceback)

    def close(self):
        self._adbc.close()



import json


def df_to_geojson(df, geom="geometry"):
    """
    Converts Pandas data frame to GeoJson-like object
    """
    return {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "geometry": json.loads(record.pop(geom)),
                "properties": record,
            }
            for record in json.loads(df.to_json(orient="records"))
        ],
    }



import adbc_driver_manager.dbapi
import pandas


class KamuConnection:
    """
    Base interface all connections implement
    """

    def url(self) -> str:
        """
        Returns the URL this connection was created with
        """
        raise NotImplementedError()

    def query(self, sql) -> pandas.DataFrame:
        """
        Execute SQL query and return result as Pandas DataFrame.
        """
        raise NotImplementedError()

    def as_adbc(self) -> adbc_driver_manager.dbapi.Connection:
        """
        Returns the underlying ADBC connection.

        Use this method when working with libraries that expect ADBC connection.

        Examples
        --------
        >>> import pandas
        >>> import kamu
        >>>
        >>> with kamu.connect() as con:
        >>>     pandas.read_sql("select 1", con.as_adbc())
        """
        raise NotImplementedError()

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def close(self):
        """
        Close the connection and release all resources
        """



import logging
import os
import re

import livy

from ._connection import KamuConnection

_logger = logging.getLogger(__package__)


def _maybe_int(value):
    if value:
        return int(value)
    return None


LIVY_DEFAULT_SESSION_PARAMS = {
    "executor_cores": _maybe_int(os.environ.get("KAMU_CLIENT_LIVY_EXECUTOR_CORES")),
    "executor_memory": os.environ.get("KAMU_CLIENT_LIVY_EXECUTOR_MEMORY"),
    "driver_cores": _maybe_int(os.environ.get("KAMU_CLIENT_LIVY_DRIVER_CORES")),
    "driver_memory": os.environ.get("KAMU_CLIENT_LIVY_DRIVER_MEMORY"),
    "heartbeat_timeout": _maybe_int(
        os.environ.get("KAMU_CLIENT_LIVY_HEARTBEAT_TIMEOUT")
    ),
}

UNRESOLVED_TABLE_RE = re.compile(r"UnresolvedRelation \[([^\]]+)\],")

SESSION_SETUP = r"""
import os

# spark.sparkContext._jvm.org.datasyslab.geosparksql.utils.GeoSparkSQLRegistrator.registerAll(sc._jvm.SQLContext(sc._jsc.sc()))

def resolve_dataset_ref(dataset_ref):
    if "/" not in dataset_ref:
        # Single-tenant
        data_path = os.path.join(dataset_ref, "data")
        if os.path.exists(data_path):
            return os.path.join(data_path, "*")
    else:
        # Multi-tenant
        # Assumptions:
        # - Layout of the data directory is `<account_name>/<dataset_id>/info/alias`
        # - Alias file contains `<account_name>/<dataset_name>`
        #   - Note there is a bug where alias may conain just `<dataset_name>` so we account for that too
        account_name, dataset_name = dataset_ref.split("/", 1)
        if os.path.isdir(account_name):
            for dataset_id in os.listdir(account_name):
                alias_path = os.path.join(account_name, dataset_id, "info", "alias")
                if not os.path.exists(alias_path):
                    continue
                with open(alias_path) as f:
                    alias = f.read().strip()
                if alias != dataset_ref and alias != dataset_name:
                    continue
                return os.path.join(account_name, dataset_id, "data", "*")

    raise Exception(f"Dataset {dataset_ref} not found")
"""


class KamuConnectionLivy(KamuConnection):
    """
    `KamuConnection` implementation using Spark Livy HTTP gateway protocol.

    This connection type is deprecated and should not be used in production. It does
    not support proper auth and has issues correctly representing certain data types.
    Livy gateway will be replaced in the near future with ADBC + FlightSQL based implementation.
    """

    def __init__(
        self, url, token=None, auto_import_datasets=True, **livy_session_params
    ):
        super().__init__()

        self._auto_import_datasets = auto_import_datasets

        livy_session_params_final = dict(LIVY_DEFAULT_SESSION_PARAMS)
        livy_session_params_final.update(livy_session_params)

        _logger.debug(
            "Creating Livy connection",
            extra={"url": url, "livy_session_params": livy_session_params_final},
        )

        self._url = url
        self._livy = livy.LivySession.create(self._url, **livy_session_params_final)
        self._livy.wait()
        self._livy.run(SESSION_SETUP)

    def __repr__(self):
        return f"KamuConnectionLivy(url='{self._url}')"

    def url(self):
        return self._url

    def query(self, sql):
        try:
            # Try to run query
            self._livy.run(f"_df = spark.sql(r'''{sql}''')")
            return self._livy.download("_df")
        except livy.models.SparkRuntimeError as err:
            # Catch "table does not exist" errors
            if not self._auto_import_datasets or not self._is_table_not_found_error(
                err
            ):
                raise
            datasets = self._parse_unresolved_tables(err)

        # Attempt to import dataset corresponding to missing table names
        _logger.debug(
            "Attempt to import dataset corresponding to missing table names",
            extra={"datasets": datasets},
        )

        for dataset in datasets:
            self._import_dataset(dataset)

        # Re-run the original query
        self._livy.run(f"_df = spark.sql(r'''{sql}''')")
        return self._livy.download("_df")

    def as_adbc(self):
        return RuntimeError(
            "Spark engine connection does not yet support ADBC client interface"
        )

    def __enter__(self):
        self._livy.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._livy.__exit__(exc_type, exc_value, traceback)

    def close(self):
        self._livy.close()

    def _is_table_not_found_error(self, err):
        return (
            err.ename == "AnalysisException"
            and "[TABLE_OR_VIEW_NOT_FOUND]" in err.evalue
        )

    def _parse_unresolved_tables(self, err):
        return [m.group(1) for m in UNRESOLVED_TABLE_RE.finditer(err.evalue)]

    def _import_dataset(self, name):
        self._livy.run(
            f"spark.read.parquet(resolve_dataset_ref('{name}')).createOrReplaceTempView('`{name}`')"
        )



import os

from ._connection import KamuConnection

__version__ = "0.6.1"


def connect(
    url=None, token=None, engine=None, connection_params=None
) -> KamuConnection:
    """
    Open connection to a Kamu node.

    Arguments
    ---------
    - `url` - URL to connect to (e.g. `grpc+tls://kamu-node.example.com`)
    - `token` - access token
    - `engine` - engine type (e.g. `datafusion`, `spark`)
    - `connection_params` - parameters to pass to the underlying connection type

    Examples
    --------
    >>> import kamu
    >>>
    >>> # Prefer using in a context manager
    >>> with kamu.connect("...") as con:
    >>>     pass
    >>>
    >>> # Connect to secure node
    >>> kamu.connect("grpc+tls://node.demo.kamu.dev:50050")
    >>>
    >>> # Connect to local insecure node
    >>> kamu.connect("grpc://localhost:50050")
    >>>
    >>> # Connect to Spark engine
    >>> kamu.connect("http://livy-host:8998", engine="spark")
    >>>
    >>> # Start a local SQL server in a workspace and connect to it:
    >>> kamu.connect("file:///path/to/workspace")
    """
    url = url or os.environ.get("KAMU_CLIENT_URL")
    if not url:
        raise ValueError("url is not specified")

    token = token or os.environ.get("KAMU_CLIENT_TOKEN")

    engine = (engine or os.environ.get("KAMU_CLIENT_ENGINE", "datafusion")).lower()

    connection_params = connection_params or {}
    connection_params["token"] = token

    if url.startswith("file://"):
        from . import _connection_subprocess

        return _connection_subprocess.KamuConnectionSubprocess(
            url=url, connection_factory=connect, engine=engine, **connection_params
        )

    if engine == "datafusion":
        from . import _connection_flight_sql

        if not url.startswith("grpc"):
            raise ValueError(
                "DataFusion engine expects URLs with 'grpc://' or 'grpc+tls://' schemes. "
                "If you are seeing this message when running `kamu notebook` CLI command "
                "- restart the notebook with `--engine datafusion` argument."
            )

        return _connection_flight_sql.KamuConnectionFlightSql(
            url=url, **connection_params
        )
    if engine == "spark":
        from . import _connection_livy

        if not url.startswith("http"):
            raise ValueError(
                "Spark engine expects a Livy HTTP server URL with 'http://' or 'https://' "
                "schemes. If you are seeing this message when running `kamu notebook` CLI command "
                "- restart the notebook with `--engine spark` argument."
            )

        return _connection_livy.KamuConnectionLivy(url=url, **connection_params)

    raise ValueError(f"Engine '{engine}' is not supported")


def load_ipython_extension(ipython):
    """
    Called when running `%load_ext kamu` in Jupyter / IPython.
    """
    from . import _jupyter

    ipython.register_magics(_jupyter.KamuMagics)

    try:
        import autovizwidget.widget.utils

        autoviz = autovizwidget.widget.utils.display_dataframe
    except ImportError:
        autoviz = None

    if autoviz:
        ipython.display_formatter.ipython_display_formatter.for_type_by_name(
            "pandas.core.frame",
            "DataFrame",
            autoviz,
        )



import pathlib

import pandas
import pytest

import kamu


def test_repr(workspace_st):
    with kamu.connect(workspace_st.url) as con:
        assert (
            repr(con)
            == f"KamuConnectionSubprocess(url='{workspace_st.url}, inner={repr(con.inner_connection())}')"
        )


def test_sql_query_minimal_datafusion(workspace_st):
    with kamu.connect(workspace_st.url, engine="datafusion") as con:
        actual = con.query("select 1 as value")
        expected = pandas.DataFrame({"value": [1]})
        pandas.testing.assert_frame_equal(expected, actual)


def test_sql_query_minimal_spark(workspace_st):
    with kamu.connect(workspace_st.url, engine="spark") as con:
        actual = con.query("select 1 as value")
        expected = pandas.DataFrame({"value": [1]})
        pandas.testing.assert_frame_equal(expected, actual)


def test_error_cli_not_installed(workspace_st):
    with pytest.raises(FileNotFoundError, match="Kamu CLI binary 'kamu-zzz' not found"):
        kamu.connect(workspace_st.url, connection_params=dict(kamu_binary="kamu-zzz"))


def test_error_workspace_not_found(tempdir):
    url = pathlib.Path(tempdir).as_uri()
    with pytest.raises(FileNotFoundError, match="Kamu workspace not found at"):
        kamu.connect(url)



import subprocess

import pandas
import pytest

import kamu
from kamu._connection_subprocess import KamuSqlServerProcess

from .conftest import Server


@pytest.fixture
def server_livy_st(workspace_st):
    proc = KamuSqlServerProcess(cwd=workspace_st.path, engine="spark")
    url = f"http://127.0.0.1:{proc.port()}"
    yield Server(port=proc.port(), url=url, workspace=workspace_st)
    proc.stop()


@pytest.fixture
def server_livy_mt(workspace_mt):
    proc = KamuSqlServerProcess(cwd=workspace_mt.path, engine="spark")
    url = f"http://127.0.0.1:{proc.port()}"
    yield Server(port=proc.port(), url=url, workspace=workspace_mt)
    proc.stop()


def pull_test_data(cwd, account=None):
    account = "" if not account else f"--account {account}"

    subprocess.run(
        f"kamu {account} pull odf+https://node.demo.kamu.dev/kamu/covid19.british-columbia.case-details.hm",
        cwd=cwd,
        shell=True,
        capture_output=True,
        check=True,
    )

    subprocess.run(
        f"kamu {account} pull odf+https://node.demo.kamu.dev/kamu/covid19.alberta.case-details.hm",
        cwd=cwd,
        shell=True,
        capture_output=True,
        check=True,
    )


def test_repr(server_livy_st):
    with kamu.connect(server_livy_st.url, engine="spark") as con:
        assert repr(con) == f"KamuConnectionLivy(url='{server_livy_st.url}')"


def test_url_scheme_check():
    with pytest.raises(ValueError):
        kamu.connect("grpc+tls://example.com", engine="spark")


def test_query_st(server_livy_st):
    pull_test_data(server_livy_st.workspace.path)

    with kamu.connect(server_livy_st.url, engine="spark") as con:
        actual = con.query(
            """
            select
                offset,
                op,
                reported_date,
                id,
                gender,
                age_group,
                location
            from `covid19.british-columbia.case-details.hm`
            order by offset
            limit 1
            """
        )

        expected = pandas.DataFrame(
            {
                "offset": [0],
                "op": [0],
                "reported_date": ["2020-01-29T00:00:00.000Z"],
                "id": [1],
                "gender": ["M"],
                "age_group": ["40s"],
                "location": ["Out of Canada"],
            }
        ).astype(
            dtype={
                "offset": "int64",
                # TODO: should be int32
                "op": "int64",
                # TODO: should be datetime64[ms, UTC]
                "reported_date": "object",
                "id": "int64",
            }
        )

        pandas.testing.assert_frame_equal(expected, actual)


def test_query_mt(server_livy_mt):
    pull_test_data(server_livy_mt.workspace.path, account="kamu")

    with kamu.connect(server_livy_mt.url, engine="spark") as con:
        actual = con.query(
            """
            select * from (
                (
                    select
                        offset,
                        op,
                        reported_date,
                        id,
                        gender
                    from `kamu/covid19.alberta.case-details.hm`
                    order by offset
                    limit 1
                )
                union all
                (
                    select
                        offset,
                        op,
                        reported_date,
                        id,
                        gender
                    from `kamu/covid19.british-columbia.case-details.hm`
                    order by offset
                    limit 1
                )
            )
            order by reported_date
            """
        )

        expected = pandas.DataFrame(
            {
                "offset": [0, 0],
                "op": [0, 0],
                "reported_date": [
                    "2020-01-29T00:00:00.000Z",
                    "2020-03-05T00:00:00.000Z",
                ],
                "id": [1, 505748],
                "gender": ["M", "F"],
            }
        ).astype(
            dtype={
                "offset": "int64",
                # TODO: should be int32
                "op": "int64",
                # TODO: should be datetime64[ms, UTC]
                "reported_date": "object",
                "id": "int64",
            }
        )

        pandas.testing.assert_frame_equal(expected, actual)


def test_query_gis_extensions(server_livy_mt):
    with kamu.connect(server_livy_mt.url, engine="spark") as con:
        actual = con.query(
            """
            select st_asgeojson(st_point(1, 2)) as point
            """
        )

        expected = pandas.DataFrame(
            {
                "point": ['{"type":"Point","coordinates":[1.0,2.0]}'],
            }
        )

        pandas.testing.assert_frame_equal(expected, actual)



import re
import subprocess

import adbc_driver_manager
import pandas
import pytest

import kamu
from kamu._connection_subprocess import KamuSqlServerProcess

from .conftest import Server


@pytest.fixture
def server_flightsql_st(workspace_st):
    proc = KamuSqlServerProcess(cwd=workspace_st.path, engine="datafusion")
    url = f"grpc://127.0.0.1:{proc.port()}"
    yield Server(port=proc.port(), url=url, workspace=workspace_st)
    proc.stop()


@pytest.fixture
def server_flightsql_mt(workspace_mt):
    proc = KamuSqlServerProcess(cwd=workspace_mt.path, engine="datafusion")
    url = f"grpc://127.0.0.1:{proc.port()}"
    yield Server(port=proc.port(), url=url, workspace=workspace_mt)
    proc.stop()


def pull_test_data(cwd):
    subprocess.run(
        "kamu --account kamu pull odf+https://node.demo.kamu.dev/kamu/covid19.british-columbia.case-details.hm",
        cwd=cwd,
        shell=True,
        capture_output=True,
        check=True,
    )


def test_version():
    assert re.fullmatch(
        r"\d\.\d\.\d", kamu.__version__
    ), "Version doesn't match the pattern"


def test_repr(server_flightsql_mt):
    with kamu.connect(server_flightsql_mt.url) as con:
        assert repr(con) == f"KamuConnectionFlightSql(url='{server_flightsql_mt.url}')"


def test_url_scheme_check():
    with pytest.raises(ValueError):
        kamu.connect("https://example.com")


def test_sql_query_minimal_st(server_flightsql_st):
    with kamu.connect(server_flightsql_st.url) as con:
        actual = con.query("select 1 as value")
        expected = pandas.DataFrame({"value": [1]})
        pandas.testing.assert_frame_equal(expected, actual)


def test_sql_query_minimal_mt(server_flightsql_mt):
    with kamu.connect(server_flightsql_mt.url) as con:
        actual = con.query("select 1 as value")
        expected = pandas.DataFrame({"value": [1]})
        pandas.testing.assert_frame_equal(expected, actual)


def test_use_after_close(server_flightsql_mt):
    con = kamu.connect(server_flightsql_mt.url)
    actual = con.query("select 1 as value")
    expected = pandas.DataFrame({"value": [1]})
    pandas.testing.assert_frame_equal(expected, actual)
    con.close()
    with pytest.raises(adbc_driver_manager.ProgrammingError):
        con.query("select 1 as value")


def test_query_pandas_interop(server_flightsql_mt):
    with kamu.connect(server_flightsql_mt.url) as con:
        actual = pandas.read_sql_query("select 1 as value", con.as_adbc())
        expected = pandas.DataFrame({"value": [1]})
        pandas.testing.assert_frame_equal(expected, actual)


def test_query(server_flightsql_mt):
    pull_test_data(server_flightsql_mt.workspace.path)

    with kamu.connect(server_flightsql_mt.url) as con:
        actual = con.query(
            """
            select
                offset,
                op,
                reported_date,
                id,
                gender,
                age_group,
                location
            from 'kamu/covid19.british-columbia.case-details.hm'
            order by offset
            limit 1
            """
        )

        expected = pandas.DataFrame(
            {
                "offset": [0],
                "op": [0],
                "reported_date": ["2020-01-29T00:00:00.000Z"],
                "id": [1],
                "gender": ["M"],
                "age_group": ["40s"],
                "location": ["Out of Canada"],
            }
        ).astype(
            dtype={
                "offset": "int64",
                "op": "int32",
                "reported_date": "datetime64[ms, UTC]",
                "id": "int64",
            }
        )

        pandas.testing.assert_frame_equal(expected, actual)



import logging
import os
import socket
import subprocess
import tempfile
import time

from ._connection import KamuConnection

_logger = logging.getLogger(__package__)


class KamuConnectionSubprocess(KamuConnection):
    """
    This special `KamuConnection` implementation is created for `file://` URLs
    and will attempt to run local [`kamu-cli`](https://github.com/kamu-data/kamu-cli)
    binary to serve the data from local FS workspace.
    """

    def __init__(
        self,
        url,
        connection_factory,
        engine,
        kamu_binary="kamu",
        start_timeout=30,
        **connection_params,
    ):
        super().__init__()

        self._url = url

        assert url.startswith("file://")
        cwd = url[len("file://") :] or "."

        # Start and wait for subprocess
        self._proc = KamuSqlServerProcess(
            bin=kamu_binary, cwd=cwd, engine=engine, start_timeout=start_timeout
        )

        # Form new URL
        if engine == "spark":
            engine_url = f"http://localhost:{self._proc.port()}"
        else:
            engine_url = f"grpc://localhost:{self._proc.port()}"

        # Connect to subprocess
        try:
            self._con = connection_factory(
                url=engine_url, engine=engine, connection_params=connection_params
            )
        except Exception:
            self._proc.stop()
            raise

    def __repr__(self):
        return f"KamuConnectionSubprocess(url='{self._url}, inner={repr(self._con)}')"

    def url(self):
        return self._url

    def inner_connection(self) -> KamuConnection:
        return self._con

    def query(self, sql):
        return self._con.query(sql)

    def as_adbc(self):
        return self._con.as_adbc

    def __enter__(self):
        self._con.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._con.__exit__(exc_type, exc_value, traceback)
        self._proc.stop()

    def close(self):
        self._con.close()
        self._proc.stop()


class KamuSqlServerProcess:
    def __init__(self, bin="kamu", cwd=".", engine="datafusion", start_timeout=15):
        self._proc = None

        if not self.find_workspace(cwd):
            raise FileNotFoundError(f"Kamu workspace not found at {cwd}")

        self._port = self.find_free_port()

        stdout = self.get_temp_file("kamu-client-supbrocess-", ".stdout.txt")
        stderr = self.get_temp_file("kamu-client-supbrocess-", ".stderr.txt")

        cmd = [
            bin,
            "-v",
            "sql",
            "server",
            "--engine",
            engine,
            "--port",
            str(self._port),
        ]

        if engine == "spark":
            cmd.append("--livy")

        _logger.debug("Starting server", extra={"cmd": str(cmd)})

        try:
            self._proc = subprocess.Popen(
                cmd,
                cwd=cwd,
                # shell=True,
                stdout=stdout,
                stderr=stderr,
                close_fds=True,
            )
        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"Kamu CLI binary '{bin}' not found. If the tool is installed consider specifying the full path"
            ) from e

        deadline = time.time() + start_timeout
        while True:
            try:
                status = self._proc.wait(1)
                raise Exception(
                    f"Kamu failed to start with status code: {status}\n"
                    f"See logs for details:\n"
                    f"- {stdout.name}\n"
                    f"- {stderr.name}"
                )
            except subprocess.TimeoutExpired:
                pass

            if self.is_port_open(self._port):
                break

            if time.time() >= deadline:
                self._proc.terminate()
                self._proc = None
                raise Exception(
                    f"Kamu failed to start within {start_timeout} seconds\n"
                    f"See logs for details:\n"
                    f"- {stdout.name}\n"
                    f"- {stderr.name}"
                )

    def port(self):
        return self._port

    def __del__(self):
        self.stop()

    def stop(self):
        if self._proc:
            self._proc.terminate()
            self._proc.wait()
            self._proc = None

    def find_workspace(self, cwd):
        current_path = os.path.abspath(cwd)

        # Stop when reaching the root directory
        while current_path != os.path.dirname(current_path):
            ws_path = os.path.join(current_path, ".kamu")
            if os.path.isdir(ws_path):
                return ws_path
            current_path = os.path.dirname(current_path)

        return None

    def find_free_port(self):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("127.0.0.1", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            free_port = s.getsockname()[1]
            return free_port

    def get_temp_file(self, prefix, extension):
        return tempfile.NamedTemporaryFile(
            prefix=prefix, suffix=extension, delete=False
        )

    def is_port_open(self, port):
        try:
            s = socket.create_connection(
                address=("127.0.0.1", port),
                timeout=0.1,
            )
        except socket.error:
            return False

        s.settimeout(1)

        try:
            read = s.recv(1)
        except TimeoutError:
            return True

        s.close()
        if len(read):
            return True

        return False
